from __future__ import annotations

from types import SimpleNamespace
from functools import wraps
from pathlib import Path
from copy import deepcopy
from typing import Any, ClassVar

from typing_extensions import Self

from ._texture import load_texture
from .._annotations import T, AnimatedNode


class Animation:
    __slots__ = ("frames",)

    def __init__(self, animation_path: Path | str, /) -> None:
        # fmt: off
        frame_directory = (
            Path.cwd()
            .joinpath(str(animation_path))
            .iterdir()
        )
        # fmt: on
        self.frames = list(map(load_texture, frame_directory))

    def __repr__(self) -> str:
        # should never be empty, but if the programmer did it, show empty frame count
        if not self.frames:
            return f"{self.__class__.__name__}(▦!:N/A->N/A)"
        longest = 0
        shortest = 0
        tallest = 0
        lowest = 0
        # these are used as temporary variables in loop
        local_longest = 0
        local_shortest = 0
        local_tallest = 0
        local_lowest = 0
        for frame in self.frames:
            # compare all time best against best results per iteration
            if not frame:  # allow empty frame
                continue
            elif not any(frame):  # allow frame with empty lines
                continue
            local_longest = len(max(frame, key=len))
            longest = max(local_longest, longest)
            local_tallest = len(frame)
            tallest = max(local_tallest, tallest)
            local_shortest = len(min(frame, key=len))
            shortest = min(local_shortest, shortest)
            local_lowest = min(local_lowest, shortest)
        return (
            self.__class__.__name__
            + "("
            + f"▦{len(self.frames)}"
            + f":{shortest}x{lowest}"
            + f"->{longest}x{tallest}"
            + ")"
        )


class AnimationMapping(SimpleNamespace):
    def __init__(self, **animations: Animation) -> None:
        super().__init__(**animations)

    def __getattribute__(self, name: str) -> Animation:
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Animation) -> None:
        return super().__setattr__(name, value)

    def get(self, animation_name: str, default: T = None) -> Animation | T:
        return getattr(self, animation_name, default)

    def update(self, animations: dict[str, Animation]) -> None:
        for name, animation in animations.items():
            setattr(self, name, animation)


# TODO: add `.play_backwards` attribute or method
# TODO: ensure last frame was rendered before `.is_playing = False`,
#       because a loop checking if it should replay the animations will
#       reset it back to the first frame before the last one is displayed
class Animated:  # Component (mixin class)
    animated_instances: ClassVar[dict[int, AnimatedNode]] = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        Animated.animated_instances[instance.uid] = instance  # type: ignore
        if (class_animations := getattr(instance, "animations", None)) is not None:
            instance.animations = deepcopy(class_animations)
        else:
            instance.animations = AnimationMapping()

        # inject `._wrapped_update_animated()` into `.update()`
        def update_method_factory(instance: AnimatedNode, bound_update):  # noqa: ANN001 ANN202
            @wraps(bound_update)
            def new_update_method(delta: float) -> None:
                bound_update(delta)  # TODO: swap order will fix rendering??
                instance._wrapped_update_animated(delta)

            return new_update_method

        instance.update = update_method_factory(instance, instance.update)  # type: ignore
        return instance  # type: ignore

    animations: AnimationMapping
    current_animation: Animation | None = None
    is_playing: bool = False
    _frame_index: int = 0

    def with_animations(self, /, **animations: Animation) -> Self:
        self.animations.update(animations)
        return self

    def with_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ) -> Self:
        self.add_animation(animation_name, animation)
        return self

    def add_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ) -> None:
        setattr(self.animations, animation_name, animation)

    def play(self, animation_name: str, /) -> None:
        self.current_animation = self.animations.get(animation_name, None)
        self.is_playing = True
        self._frame_index = 0
        # the actual logic of playing the animation is handled in `.update(...)`

    def _wrapped_update_animated(self, _delta: float) -> None:
        if self.current_animation is None:
            self.is_playing = False
            return
        self.texture = self.current_animation.frames[self._frame_index]
        frame_count = len(self.current_animation.frames)
        self._frame_index = min(self._frame_index + 1, frame_count - 1)
        if self._frame_index == frame_count - 1:
            self.is_playing = False

    def _free(self) -> None:
        del Animated.animated_instances[self.uid]  # type: ignore
        super()._free()  # type: ignore



from __future__ import annotations

from typing import Any

from colex import ColorValue
from typing_extensions import Self

from .._annotations import ColorNode


class Color:  # Component (mixin class)
    color_instances: dict[int, ColorNode] = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        Color.color_instances[instance.uid] = instance  # type: ignore
        return instance

    color: ColorValue | None = None

    def with_color(self, color: ColorValue | None, /) -> Self:
        self.color = color
        return self

    def _free(self) -> None:
        del Color.color_instances[self.uid]  # type: ignore
        super()._free()  # type: ignore



"""
Charz
-----

An object oriented terminal game engine

Includes:
- `Engine` (derive new App class from this)
- `Clock` (calculates delta time)
- `DeltaClock` (controls framerate)
- `Screen` (renders nodes to console)
- `Camera` (handles viewport)
- `CameraMode` (enum)
- `Node` (base node)
- `Node2D` (prefabricated, node in world)
- `Transform` (component)
- `lerp` (function from `linflex` package)
- `sign` (function from `linflex` package)
- `clamp` (function from `linflex` package)
- `Vec2` (datastructure from `linflex` package)
- `Vec2i` (datastructure from `linflex` package)
- `Vec3` (datastructure from `linflex` package)
- `load_texture` (function)
- `Texture` (component)
- `Color` (component)
- `ColorValue` (annotation from `colex` package)
- `Label` (prefabricated)
- `Sprite` (prefabricated)
- `AnimatedSprite` (prefabricated)
- `Animation` (datastructure)
- `AnimationMapping` (datastructure)
- `Collider` (component)
- `Hitbox` (datastructure)
- `text` (module for flipping strings)
"""

from __future__ import annotations as _annotations

__all__ = [
    "Engine",
    "Clock",
    "DeltaClock",
    "Screen",
    "Camera",
    "CameraMode",
    "Node",
    "Node2D",
    "Transform",
    "lerp",
    "sign",
    "clamp",
    "Vec2",
    "Vec2i",
    "Vec3",
    "load_texture",
    "Texture",
    "Color",
    "ColorValue",
    "Label",
    "Sprite",
    "AnimatedSprite",
    "Animation",
    "AnimationMapping",
    "Collider",
    "Hitbox",
    "text",
]

from linflex import lerp, sign, clamp, Vec2, Vec2i, Vec3
from colex import ColorValue

from ._engine import Engine
from ._clock import Clock, DeltaClock
from ._screen import Screen
from ._camera import Camera, CameraMode
from ._node import Node
from ._components._transform import Transform
from ._components._texture import load_texture, Texture
from ._components._color import Color
from ._components._animation import Animation, AnimationMapping
from ._components._collision import Collider, Hitbox
from ._prefabs._node2d import Node2D
from ._prefabs._label import Label
from ._prefabs._sprite import Sprite
from ._prefabs._animated_sprite import AnimatedSprite
from . import text



from __future__ import annotations

from typing import Any

from typing_extensions import Self

from ._clock import Clock, DeltaClock
from ._screen import Screen
from ._node import Node


class EngineMixinSortMeta(type):
    """Engine metaclass for initializing `Engine` subclass after other `mixin` classes"""

    def __new__(cls, name: str, bases: tuple[type, ...], attrs: dict[str, object]):
        def sorter(base: type) -> bool:
            return isinstance(base, Engine)

        sorted_bases = tuple(sorted(bases, key=sorter))
        new_type = super().__new__(cls, name, sorted_bases, attrs)
        return new_type


class Engine(metaclass=EngineMixinSortMeta):
    fps: float | None = 16
    clock: Clock = DeltaClock()
    screen: Screen = Screen()
    clear_console: bool = False
    hide_cursor: bool = True
    is_running: bool = False

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        # set `.clock.tps` with `.fps` set from class attribute
        instance.clock.tps = instance.fps
        return instance  # type: ignore

    def update(self, delta: float) -> None: ...

    def run(self) -> Self:
        # check if console/stream should be cleared
        if self.clear_console:
            clear_code = "\x1b[2J\x1b[H"
            self.screen.stream.write(clear_code)
            self.screen.stream.flush()
        # hide cursor
        if self.hide_cursor:
            hide_code = "\x1b[?25l"
            self.screen.stream.write(hide_code)
            self.screen.stream.flush()

        delta = self.clock.delta()  # initial delta
        self.is_running = True

        while self.is_running:  # main loop
            self.update(delta)
            for queued_node in Node._queued_nodes:
                queued_node._free()
            Node._queued_nodes *= 0  # NOTE: faster way to do `.clear()`
            # NOTE: `list` is faster than `tuple`, when copying
            for node in list(Node.node_instances.values()):  # iterating copy
                node.update(delta)
            self.screen.refresh()
            self.clock.tick()
            delta = self.clock.delta()

        # show cursor if hidden
        if self.hide_cursor:
            hide_code = "\x1b[?25h"
            self.screen.stream.write(hide_code)
            self.screen.stream.flush()

        return self  # this is for convenience if the desire to read the app state arise
